/*
 * DoublyLinkedList unit tests
 * Adapated from https://courses.cs.washington.edu/courses/cse373/17au/project1/project1.zip
 */

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;

import java.util.NoSuchElementException;
import java.util.Random;

public class TestLinkedList {
    private static final int SECOND = 1000;

    /**
     * This method creates a simple list containing three elements to help minimize
     * redundancy later in our tests.
     */
    private DoublyLinkedList<String> makeBasicList() {
        DoublyLinkedList<String> list = new DoublyLinkedList<>();

        list.addLast("a");
        list.addLast("b");
        list.addLast("c");

        return list;
    }

    /**
     * This test will check if a list contains exactly the same elements as
     * the "expected" array.
     */
    private <T> void assertListMatches(T[] expected, DoublyLinkedList<T> actual) {
        assertEquals(expected.length, actual.size());
        assertEquals(expected.length == 0, actual.isEmpty());

        for (int i = 0; i < expected.length; i++) {
            try {
                assertEquals("Item at index " + i + " does not match", expected[i], actual.get(i));
            } catch (Exception ex) {
                String errorMessage = String.format(
                        "Got %s when getting item at index %d (expected '%s')",
                        ex.getClass().getSimpleName(),
                        i,
                        expected[i]);
                throw new AssertionError(errorMessage, ex);
            }
        }
    }

    /**
     * Note: We use 1 second as the default timeout for many of our tests.
     *
     * One second is typically extremely generous: most of your tests should
     * finish in milliseconds. If one of your tests is timing out, you're almost
     * certainly doing something wrong.
     */

    @Test
    @Timeout(SECOND)
    public void testAddAndGetBasic() {
        DoublyLinkedList<String> list = makeBasicList();
        this.assertListMatches(new String[] {"a", "b", "c"}, list);

        list = new DoublyLinkedList<>();
        list.addFirst("c");
        list.addFirst("b");
        list.addFirst("a");
        this.assertListMatches(new String[] {"a", "b", "c"}, list);
    }

    @Test
    @Timeout(SECOND)
    public void testToString() {
        DoublyLinkedList<String> list = makeBasicList();
        assertEquals("[a, b, c]", list.toString());

        list = new DoublyLinkedList<>();
        assertEquals("[]", list.toString());
        list.addFirst("x");
        assertEquals("[x]", list.toString());
    }

    @Test
    @Timeout(2 * SECOND)
    public void testAddAndGetWorksForManyNumbers() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        int cap = 1000;
        for (int i = 0; i < cap; i++) {
            list.addLast(i * 2);
        }
        assertEquals(cap, list.size());
        for (int i = 0; i < cap; i++) {
            int value = list.get(i);
            assertEquals(i* 2, value);
        }
        assertEquals(cap, list.size());
    }

    @Test
    @Timeout(15 * SECOND)
    public void testAddIsEfficient() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        int cap = 5000000;
        for (int i = 0; i < cap; i++) {
            list.addLast(i * 2);
        }
        assertEquals(cap, list.size());
    }

    @Test
    @Timeout(SECOND)
    public void testRemoveMultiple() {
        DoublyLinkedList<String> list = this.makeBasicList();
        assertEquals("c", list.removeLast());
        this.assertListMatches(new String[] {"a", "b"}, list);

        assertEquals("b", list.removeLast());
        this.assertListMatches(new String[] {"a"}, list);

        assertEquals("a", list.removeLast());
        this.assertListMatches(new String[] {}, list);

        list = this.makeBasicList();
        assertEquals("a", list.removeFirst());
        this.assertListMatches(new String[] {"b", "c"}, list);

        assertEquals("b", list.removeFirst());
        this.assertListMatches(new String[] {"c"}, list);

        assertEquals("c", list.removeFirst());
        this.assertListMatches(new String[] {}, list);
    }

    @Test
    @Timeout(SECOND)
    public void testAddAndRemoveFromEnd() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        int cap = 1000;

        for (int i = 0; i < cap; i++) {
            list.addLast(i);
        }

        assertEquals(cap, list.size());

        for (int i = cap - 1; i >= 0; i--) {
            int value = list.removeLast();
            assertEquals(i, value);
        }

        assertEquals(0, list.size());
    }

    @Test
    @Timeout(SECOND)
    public void testAlternatingAddAndRemove() {
        int iterators = 1000;

        DoublyLinkedList<String> list = new DoublyLinkedList<>();

        for (int i = 0; i < iterators; i++) {
            String entry = "" + i;
            list.addLast(entry);
            assertEquals(1, list.size());

            String out = list.removeLast();
            assertEquals(entry, out);
            assertEquals(0, list.size());
        }
    }

    @Test
    @Timeout(5 * SECOND)
    public void testAddAndRemoveFromEndIsEfficient() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        for (int i = 0; i < 10000; i++) {
            list.addLast(i);
        }

        for (int i = 0; i < 10000; i++) {
            list.addLast(-1);
            list.removeLast();
        }
    }

    @Test
    @Timeout(SECOND)
    public void testRemoveOnEmptyListThrowsException() {
        DoublyLinkedList<String> list = this.makeBasicList();
        list.removeLast();
        list.removeLast();
        list.removeLast();
        try {
            list.removeLast();
            // We didn't throw an exception? Fail now.
            fail("Expected NoSuchElementException");
        } catch (NoSuchElementException ex) {
            // Do nothing: this is ok
        }
    }

    @Test
    @Timeout(SECOND)
    public void testGetOutOfBoundsThrowsException() {
        DoublyLinkedList<String> list = this.makeBasicList();
        try {
            list.get(-1);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }

        // This should be ok
        list.get(2);

        try {
            // Now we're out of bounds
            list.get(3);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }

        try {
            list.get(1000);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }
    }

    @Test
    @Timeout(SECOND)
    public void testSetElements() {
        DoublyLinkedList<String> list = this.makeBasicList();

        list.set(0, "AAA");
        assertListMatches(new String[] {"AAA", "b", "c"}, list);

        list.set(1, "BBB");
        assertListMatches(new String[] {"AAA", "BBB", "c"}, list);

        list.set(2, "CCC");
        assertListMatches(new String[] {"AAA", "BBB", "CCC"}, list);
    }

    @Test
    @Timeout(SECOND)
    public void testSetWithOneElement() {
        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        list.addLast("foo");

        list.set(0, "bar");
        assertListMatches(new String[] {"bar"}, list);

        list.set(0, "baz");
        assertListMatches(new String[] {"baz"}, list);
    }

    @Test
    @Timeout(SECOND)
    public void testSetOutOfBoundsThrowsException() {
        DoublyLinkedList<String> list = this.makeBasicList();

        try {
            list.set(-1, "AAA");
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // This is ok: do nothing
        }

        try {
            list.set(3, "AAA");
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // This is ok: do nothing
        }
    }

    @Test
    @Timeout(5 * SECOND)
    public void testSetManyItems() {
        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        int cap = 10000;

        for (int i = 0; i < cap; i++) {
            list.addLast("foo" + i);
        }

        for (int i = 0; i < cap; i++) {
            list.set(i, "bar" + i);
        }

        for (int i = 0; i < cap; i++) {
            assertEquals("bar" + i, list.get(i));
        }

        for (int i = cap - 1; i >= 0; i--) {
            list.set(i, "qux" + i);
        }

        for (int i = cap - 1; i >= 0; i--) {
            assertEquals("qux" + i, list.get(i));
        }
    }

    @Test
    @Timeout(SECOND)
    public void testRemoveBasic() {
        DoublyLinkedList<String> list = this.makeBasicList();
        String ret = list.remove(0);
        this.assertListMatches(new String[] {"b", "c"}, list);
        assertEquals("a", ret);

        ret = list.remove(1);
        this.assertListMatches(new String[] {"b"}, list);
        assertEquals("c", ret);
    }

    @Test
    @Timeout(SECOND)
    public void testRemoveElementBasic() {
        DoublyLinkedList<String> list = this.makeBasicList();
        String ret = list.remove("b");
        this.assertListMatches(new String[] {"a", "c"}, list);
        assertEquals("b", ret);

        ret = list.remove("a");
        this.assertListMatches(new String[] {"c"}, list);
        assertEquals("a", ret);

        ret = list.remove("c");
        this.assertListMatches(new String[] {}, list);
        assertEquals("c", ret);
    }

    @Test
    @Timeout(SECOND)
    public void testRemoveEmptyAndSingleElement() {
        // Lists 1 and 2: insert into empty
        DoublyLinkedList<String> list1 = new DoublyLinkedList<>();
        DoublyLinkedList<String> list2 = new DoublyLinkedList<>();
        list1.addFirst("a");

        this.assertListMatches(new String[] {"a"}, list1);

        // List 1: remove single element
        list1.remove(0);
        this.assertListMatches(new String[] {}, list1);

        // List 2: remove empty
        try {
            list2.remove(0);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // do nothing: this is ok
        }
    }

    @Test
    public void testRemoveOutOfBounds() {
        DoublyLinkedList<String> list = this.makeBasicList();

        try {
            list.remove(-1);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }

        try {
            list.remove(4);
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }
    }

    @Test
    @Timeout(15 * SECOND)
    public void testRemoveAtFrontIsEfficient() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        int cap = 5000000;
        for (int i = 0; i < cap; i++) {
            list.addFirst(i * 2);
        }
        assertEquals(cap, list.size());
        for (int i = 0; i < cap; i++) {
            list.remove(0);
        }
        assertEquals(0, list.size());
    }

    @Test
    @Timeout(SECOND)
    public void testInsertBasic() {
        DoublyLinkedList<String> list = this.makeBasicList();
        list.add(0, "x");
        this.assertListMatches(new String[] {"x", "a", "b", "c"}, list);

        list.add(2, "y");
        this.assertListMatches(new String[] {"x", "a", "y", "b", "c"}, list);

        list.add(5, "z");
        this.assertListMatches(new String[] {"x", "a", "y", "b", "c", "z"}, list);
    }

    @Test
    @Timeout(SECOND)
    public void testInsertEmptyAndSingleElement() {
        // Lists 1 and 2: insert into empty
        DoublyLinkedList<String> list1 = new DoublyLinkedList<>();
        DoublyLinkedList<String> list2 = new DoublyLinkedList<>();
        list1.add(0, "a");
        list2.add(0, "a");

        // No point in checking both lists
        this.assertListMatches(new String[] {"a"}, list1);

        // List 1: insert at front
        list1.add(0, "b");
        this.assertListMatches(new String[] {"b", "a"}, list1);

        // List 2: insert at end
        list2.add(1, "b");
        this.assertListMatches(new String[] {"a", "b"}, list2);
    }

    @Test
    public void testInsertOutOfBounds() {
        DoublyLinkedList<String> list = this.makeBasicList();

        try {
            list.add(-1, "a");
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }

        try {
            list.add(4, "a");
            fail("Expected IndexOutOfBoundsException");
        } catch (IndexOutOfBoundsException ex) {
            // Do nothing: this is ok
        }
    }

    // @Test
    // @Timeout(15 * SECOND)
    // public void testInsertAtEndIsEfficient() {
    //     DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
    //     int cap = 5000000;
    //     for (int i = 0; i < cap; i++) {
    //         list.add(list.size(), i * 2);
    //     }
    //     assertEquals(cap, list.size());
    // }

    // @Test
    // @Timeout(15 * SECOND)
    // public void testInsertNearEndIsEfficient() {
    //     LinkedList<Integer> list = new DoublyLinkedList<>();
    //     list.addLast(-1);
    //     list.addLast(-2);

    //     int cap = 5000000;
    //     for (int i = 0; i < cap; i++) {
    //         list.add(list.size() - 2, i * 2);
    //     }
    //     assertEquals(cap + 2, list.size());
    // }

    @Test
    @Timeout(15 * SECOND)
    public void testInsertAtFrontIsEfficient() {
        DoublyLinkedList<Integer> list = new DoublyLinkedList<>();
        int cap = 5000000;
        for (int i = 0; i < cap; i++) {
            list.addFirst(i * 2);
        }
        assertEquals(cap, list.size());
    }

    @Test
    @Timeout(SECOND)
    public void testIndexOfAndContainsBasic() {
        DoublyLinkedList<String> list = new DoublyLinkedList<>();

        list.addLast("a");
        list.addLast("b");
        list.addLast("c");
        list.addLast("q");
        list.addLast("a");

        // Test index of
        assertEquals(0, list.indexOf("a"));
        assertEquals(2, list.indexOf("c"));
        assertEquals(-1, list.indexOf("z"));

        // Test equivalent logic using list.contains(...)
        assertTrue(list.contains("a"));
        assertTrue(list.contains("c"));
        assertFalse(list.contains("z"));
    }

    @Test
    @Timeout(SECOND)
    public void testIndexOfAndContainsCorrectlyComparesItems() {
        // Two different String objects, but with equal values
        String item1 = "abcdefghijklmnopqrstuvwxyz";
        String item2 = item1 + "";

        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        list.addLast("foo");
        list.addLast(item1);

        assertEquals(1, list.indexOf(item2));
        assertTrue(list.contains(item2));
    }

    @Test
    @Timeout(5 * SECOND)
    public void testIndexOfAndContainsMany() {
        int cap = 1000;
        int stringLength = 100;
        String validChars = "abcdefghijklmnopqrstuvwxyz0123456789";

        // By setting the seed to some arbitrary but constant number, we guarantee
        // this random number generator will produce the exact same sequence of numbers
        // every time we run this test. This helps us keep our tests deterministic, which
        // can help with debugging.
        Random rand = new Random();
        rand.setSeed(12345);

        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        DoublyLinkedList<String> refList = new DoublyLinkedList<>();

        for (int i = 0; i < cap; i++) {
            String entry = "";
            for (int j = 0; j < stringLength; j++) {
                int charIndex = rand.nextInt(validChars.length());
                entry += validChars.charAt(charIndex);
            }

            list.addLast(entry);
            if (i % 100 == 0) {
                refList.addLast(entry);
            }
        }

        for (int i = 0; i < refList.size(); i++) {
            String entry = refList.get(i);
            assertEquals(i * 100, list.indexOf(entry));
            assertTrue(list.contains(entry));
        }
    }

    @Test
    @Timeout(SECOND)
    public void testLastIndexOfBasic() {
        DoublyLinkedList<String> list = new DoublyLinkedList<>();

        list.addLast("a");
        list.addLast("b");
        list.addLast("c");
        list.addLast("q");
        list.addLast("a");

        // Test index of
        assertEquals(4, list.lastIndexOf("a"));
        assertEquals(2, list.lastIndexOf("c"));
        assertEquals(-1, list.lastIndexOf("z"));
    }

    @Test
    @Timeout(SECOND)
    public void testLastIndexOfCorrectlyComparesItems() {
        // Two different String objects, but with equal values
        String item1 = "abcdefghijklmnopqrstuvwxyz";
        String item2 = item1 + "";

        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        list.addLast("foo");
        list.addLast(item1);

        assertEquals(1, list.lastIndexOf(item2));
    }

    @Test
    @Timeout(5 * SECOND)
    public void testLastIndexOfMany() {
        int cap = 1000;
        int stringLength = 100;
        String validChars = "abcdefghijklmnopqrstuvwxyz0123456789";

        // By setting the seed to some arbitrary but constant number, we guarantee
        // this random number generator will produce the exact same sequence of numbers
        // every time we run this test. This helps us keep our tests deterministic, which
        // can help with debugging.
        Random rand = new Random();
        rand.setSeed(12345);

        DoublyLinkedList<String> list = new DoublyLinkedList<>();
        DoublyLinkedList<String> refList = new DoublyLinkedList<>();

        for (int i = 0; i < cap; i++) {
            String entry = "";
            for (int j = 0; j < stringLength; j++) {
                int charIndex = rand.nextInt(validChars.length());
                entry += validChars.charAt(charIndex);
            }

            list.addLast(entry);
            if (i % 100 == 0) {
                refList.addLast(entry);
            }
        }

        for (int i = 0; i < refList.size(); i++) {
            String entry = refList.get(i);
            assertEquals(i * 100, list.lastIndexOf(entry));
        }
    }
}
